from nose.tools import eq_

from ..apply import apply
from ..tokenizers import text_split


def diff_and_replay(diff):
    a = """
    This sentence is going to get copied. This sentence is going to go away.

    This is a paragraph that is mostly going to change.  However, there's going
    to be a sentence right in the middle that stays.  And now we're done with
    that.

    This is another sentence. asldknasl dsal dals dals dlasd oa kdlawbndkubawdk
    """

    b = """
    This sentence is going to get copied.  Wha... a new thing appeared!

    Everyone thought that this paragraph would totally change.  However, there's
    going to be a sentence right in the middle that stays.  Isn't that funny!?

    This is another sentence. This sentence is going to get copied.
    """
    a_tokens = list(text_split.tokenize(a))
    b_tokens = list(text_split.tokenize(b))
    operations = list(diff(a_tokens, b_tokens))

    print("Diff 1:")
    for op in operations:
        if op.name == "equal":
            print("equal: " + repr("".join(a_tokens[op.a1:op.a2])))
        elif op.name == "delete":
            print("delete: " + repr("".join(a_tokens[op.a1:op.a2])))
        elif op.name == "insert":
            print("insert: " + repr("".join(b_tokens[op.b1:op.b2])))

    replay_b = [str(t) for t in apply(operations, a_tokens, b_tokens)]
    eq_(b, ''.join(replay_b))

    a = "I'm new here.  This sentence is a sentence. I'm new here."
    b = "I'm new here. Sentence is a sentence."

    a_tokens = list(text_split.tokenize(a))
    b_tokens = list(text_split.tokenize(b))
    operations = list(diff(a_tokens, b_tokens))

    print("\nDiff 2:")
    for op in operations:
        if op.name == "equal":
            print("equal: " + repr("".join(a_tokens[op.a1:op.a2])))
        elif op.name == "delete":
            print("delete: " + repr("".join(a_tokens[op.a1:op.a2])))
        elif op.name == "insert":
            print("insert: " + repr("".join(b_tokens[op.b1:op.b2])))

    replay_b = [str(t) for t in apply(operations, a_tokens, b_tokens)]
    eq_(b, ''.join(replay_b))

    a = """This is a test paragraph.  It has some sentences.

    I have a lovely bunch of coconuts.

    This is another test paragraph.  It also has some sentences.

    This is a test sentence just floating in space."""

    b = """This is a test paragraph.  It has some sentences.

    This is another test paragraph.  It also has some sentences.

    I have a lovely bunch of coconuts.

    This is a test sentence just floating in space."""

    a_tokens = list(text_split.tokenize(a))
    b_tokens = list(text_split.tokenize(b))
    operations = list(diff(a_tokens, b_tokens))

    print("\nDiff 3:")
    for op in operations:
        if op.name == "equal":
            print("equal: " + repr("".join(a_tokens[op.a1:op.a2])))
        elif op.name == "delete":
            print("delete: " + repr("".join(a_tokens[op.a1:op.a2])))
        elif op.name == "insert":
            print("insert: " + repr("".join(b_tokens[op.b1:op.b2])))

    replay_b = [str(t) for t in apply(operations, a_tokens, b_tokens)]
    eq_(b, ''.join(replay_b))



from .apply import apply
from .operations import Operation, Insert, Delete, Equal
from .algorithms.diff_engine import DiffEngine
from .algorithms import segment_matcher, SegmentMatcher
from .algorithms import sequence_matcher, SequenceMatcher
from .tokenizers import (Token, Tokenizer, RegexTokenizer, text_split,
                         wikitext_split, wikitext_split_w_cjk)
from .segmenters import (Segmenter, Segment, MatchableSegment,
                         ParagraphsSentencesAndWhitespace)

from .about import (__name__, __version__, __author__, __author_email__,
                    __description__, __license__, __url__)

__all__ = [apply,
           Operation, Insert, Delete, Equal,
           DiffEngine,
           segment_matcher, SegmentMatcher,
           sequence_matcher, SequenceMatcher,
           Token, Tokenizer, RegexTokenizer, text_split, wikitext_split,
           wikitext_split_w_cjk, Segmenter, Segment, MatchableSegment,
           ParagraphsSentencesAndWhitespace,
           __name__, __version__, __author__, __author_email__,
           __description__, __license__, __url__]




from .tokenizer import TokenizerPipeline
from .tokenizer import CJKProcessor
from .wikitext_split import wikitext_split
from . import lexicon_chars

CJK_LEXICON = {
    'cjk': r'['+lexicon_chars.cjk+r']',
    'japanese': r'['+lexicon_chars.jap_char+r']',
    'korean': r'['+lexicon_chars.kor_char+r']',
}

wikitext_split_w_cjk = TokenizerPipeline(wikitext_split,
                                         CJKProcessor(CJK_LEXICON))



from nose.tools import eq_

from ...operations import Delete, Equal, Insert
from ...tests.diff_and_replay import diff_and_replay
from ...tests.diff_sequence import diff_sequence
from ...tokenizers import wikitext_split
from ..segment_matcher import diff, process


def test_diff_and_replay():
    return diff_and_replay(diff)


def test_engine():
    return diff_sequence(process)


def test_easy_diff():
    a = "Apples are red."
    b = "Apples are tasty and red."

    operation_tokens = process([a, b], tokenizer=wikitext_split)

    # Apples are red.
    next(operation_tokens)

    # Apples are tasty and red.
    operations, a, b = next(operation_tokens)

    eq_(
        list(operations),
        [
            Equal(0, 4, 0, 4),
            Insert(4, 4, 4, 8),
            Equal(4, 6, 8, 10)
        ]
    )


def test_sentence_sandwich():
    a = """==Senior Tours==

Golfers over the age of fifty are eligible to compete in senior touraments.
[[Golf]] is unique among [[sports]] in having high profile and lucrative
competitions for this age group. Nearly all of the famous golfers who are
eligible to compete in these events choose to do so these events, unless
they are unable to for health reasons. The two main tours are:

*[[Champions Tour]] (based in the [[United States]]}
*[[European Seniors Tour]]"""

    b = """==Senior Tours==

Golfers over the age of fifty are eligible to compete in senior touraments.
[[Golf]] is unique among [[sports]] in having high profile and lucrative
competitions for this age group. Nearly all of the famous golfers who are
eligible to compete in these events choose to do so, unless
they are unable to for health reasons. The two main tours are:

*[[Champions Tour]] (based in the [[United States]]}
*[[European Seniors Tour]]"""

    operation_tokens = process([a, a, b], tokenizer=wikitext_split)

    operations, a, b = next(operation_tokens)

    operations, a, b = next(operation_tokens)

    operations, a, b = next(operation_tokens)

    eq_(
        list(operations),
        [Equal(name='equal', a1=0, a2=105, b1=0, b2=105),
         Delete(name='delete', a1=105, a2=109, b1=105, b2=105),
         Equal(name='equal', a1=109, a2=168, b1=105, b2=164)]
    )


def test_revisions():
    from ...segmenters import ParagraphsSentencesAndWhitespace
    ParagraphsSentencesAndWhitespace()
    a = """
    {| class=&quot;wikitable&quot; |}
    #&quot;Huger than Huge&quot; – ''Jordan''
    """  # noqa

    b = """
    {| class=&quot;wikitable&quot; |}
    #&quot;Huger than Huge&quot; – ''Jordan (of Dan and Jordan)''
    """  # noqa
    at = wikitext_split.tokenize(a)
    bt = wikitext_split.tokenize(b)
    operations = diff(at, bt)

    added_content = ", ".join("".join(bt[i] for i in range(op.b1, op.b2))
                              for op in operations if op.name == "insert")
    eq_(added_content, " (of Dan and Jordan)")



from nose.tools import eq_

from ..text_split import text_split


def test_simple_text_split():

    input = "As a sentence, this includes punctuation. \n" + \
            "\n" + \
            "And then we have another sentence here!"

    expected = ['As', ' ', 'a', ' ', 'sentence', ',', ' ', 'this', ' ',
                'includes', ' ', 'punctuation', '.', ' \n\n', 'And', ' ',
                'then', ' ', 'we', ' ', 'have', ' ', 'another', ' ', 'sentence',
                ' ', 'here', '!']

    tokens = list(text_split.tokenize(input))

    eq_(tokens, expected)




class LookAhead:

    class DONE:
        def __len__(self): return 0

    def __new__(cls, it):
        if isinstance(it, cls):
            return it
        elif hasattr(it, "__next__") or hasattr(it, "__iter__"):
            return cls.from_iterable(it)
        else:
            raise TypeError("Expected iterable, got {0}", type(it))

    @classmethod
    def from_iterable(cls, iterable):
        instance = super().__new__(cls)
        instance.initialize(iterable)
        return instance

    def __init__(self, *args, **kwargs): pass

    def initialize(self, iterable):
        self.iterable = iter(iterable)
        self.i = -1  # Will increment to zero in a moment
        self._load_next()

    def _load_next(self):
        try:
            self.next = next(self.iterable)
            self.i += 1
        except StopIteration:
            self.next = self.DONE

    def __iter__(self): return self

    def __next__(self):
        if self.empty():
            raise StopIteration()
        else:
            current = self.next
            self._load_next()
            return current

    def pop(self):
        return self.__next__()

    def peek(self, default=DONE):
        if not self.empty():
            return self.next
        else:
            return default

    def empty(self):
        return self.next == self.DONE



"""
A :class:`deltas.DiffEngine` implements a streaming diff strategy that applies
a diff algorithm to a sequence of text revisions.  By maintaining an internal
state, various efficiencies can be implemented.

.. autoclass:: deltas.DiffEngine
    :members:
"""
import yamlconf


class DiffEngine:
    """
    Constructs a diff engine.
    """

    class Processor:
        """
        Constructs a new diff processor for processing many versions
        of a single text.
        """

        def process(text, token_class=None):
            raise NotImplementedError()

    def processor(self):
        """
        Configures and returns a new :class:`~deltas.DiffEngine.Processor`
        """
        raise NotImplementedError()

    @classmethod
    def from_config(cls, config, name, section_key="diff_engines"):
        """
        Constructs a :class:`deltas.DiffEngine` from a configuration
        doc.
        """
        section = config[section_key][name]
        if 'module' in section:
            return yamlconf.import_module(section['module'])
        else:
            Engine = yamlconf.import_module(section['class'])
            return Engine.from_config(config, name, section_key=section_key)



from nose.tools import eq_

from ..apply import apply
from ..operations import Insert, Equal, Delete


def test_apply():
    a = [0, 1, 2, 3, 4, 5, 6]
    b = [0, 1, 2, 3, "four", "five", "six", "seven", "eight", "nine", 6]
    replayed_b = list(apply([Insert(0, 0, 0, 4), Delete(4, 6, 4, 4),
                             Insert(4, 4, 4, 10), Equal(6, 7, 10, 11)],
                            a, b))
    eq_(b, replayed_b)



from .tokenizer import RegexTokenizer
from . import lexicon_chars

PLAIN_PROTO = [r'bitcoin', r'geo', r'magnet', r'mailto', r'news', r'sips?',
               r'tel', r'urn']
SLASHED_PROTO = [r'https?', r'ftp', r'ftps', r'git', r'gopher', r'ircs?',
                 r'mms', r'nntp', r'redis', r'sftp', r'ssh', r'svn', r'telnet',
                 r'worldwind', r'xmpp']
ADDRESS = r'[^\s/$.?#][^|<>\s{}]*'

url = (
    r'(?:' +  # noqa
        r'(?:' + '|'.join(PLAIN_PROTO) + r')\:|' +  # noqa
        r'(?:(?:' + '|'.join(SLASHED_PROTO) + r')\:)?\/\/' +
    r')' + ADDRESS
)

combined_word = lexicon_chars.devangari_word + \
                lexicon_chars.arabic_word + \
                lexicon_chars.bengali_word

# ?     ab? will match either ‘a’ or ‘ab’
# []	Used to indicate a set of characters. In a set
# ^	    If the first character of the set is '^', all the
#       characters that are not in the set will be matched. For example,
#       [^5] will match any character except '5'
# \w	matches any alphanumeric character and the underscore; this is
#       equivalent to the set [a-zA-Z0-9_]
# \W	matches any non-alphanumeric character; this is equivalent to
#       the set [^a-zA-Z0-9_]
# \d	matches any decimal digit; this is equivalent to the set [0-9]
# |	    or
# \s	matches any whitespace character, this is equivalent to the
#       set [ \t\n\r\f\v]
# $	    Matches the end of the string or just before the newline at the
#       end of the string
# *     match 0 or more repetitions of the preceding RE, as many repetitions as
#       are possible. ab* will match ‘a’, ‘ab’, or ‘a’ followed by any number
#       of ‘b’s.

# (anything that is not(non-alphanmeric, decimal digit) or word_char) +
# any number of word_char repetitions +
# any number of word_char repetitions encapsulated in apostrophes

word = r'(?:[^\W\d]|[' + combined_word + r'])' + \
       r'[\w' + combined_word + r']*' + \
       r'(?:[\'’](?:[\w' + combined_word + r']+|(?=(?:$|\s))))*'

cjk_word = r'(?=[' + lexicon_chars.cjk + r'])' + \
       r'[' + lexicon_chars.cjk + r']*' + \
       r'(?:[\'’](?:[' + lexicon_chars.cjk + r']+|(?=(?:$|\s))))*'

LEXICON = [
    ('break', r'(?:\n\r?|\r\n)\s*(?:\n\r?|\r\n)+'),
    ('whitespace', r'(?:\n\r?|[^\S\n\r]+)'),
    ("url", url),
    ("equals", r"=+"),
    ("bar", r"\|"),
    ('entity', r'&[a-z][a-z0-9]*;'),
    ('ref_open', r'<ref\b(?:\/(?!>)|[^>\/])*>'),
    ('ref_close', r'</ref\b[^>]*>'),
    ('ref_singleton', r'<ref\b(?:\/(?!>)|[^/>])*\/>'),
    ('tag', r'</?([a-z][a-z0-9]*)\b[^>]*>'),
    ('number', r'\d+'),
    ("bold", r"'''"),
    ("italic", r"''"),
    ('japan_punct', r'[\u3000-\u303F]'),
    ('cjk_word', cjk_word),
    ('word', word),
    ('tab_open', r'\{\|'),
    ('tab_close', r'\|\}'),
    ('dbrack_open', r'\[\['),
    ('dbrack_close', r'\]\]'),
    ('brack_open', r'\['),
    ('brack_close', r'\]'),
    ('paren_open', r'\('),
    ('paren_close', r'\)'),
    ('dcurly_open', r'\{\{'),
    ('dcurly_close', r'\}\}'),
    ('curly_open', r'\{'),
    ('curly_close', r'\}'),
    ('period', r'\.+'),
    ('qmark', r'\?+'),
    ('epoint', r'!+'),
    ('comma', r',+'),
    ('colon', r':+'),
    ('scolon', r';+'),
    ('comment_start', r'<!--'),
    ('comment_end', r'-->'),
    ('danda', r'।|॥'),
    ("etc", r"."),
]

wikitext_split = RegexTokenizer(LEXICON)



from .segments import Segment


def print_tree(segment, depth=0):
    if isinstance(segment, Segment):
        print(depth * "\t" + "{0}: '{1}'"
                             .format(segment.__class__.__name__, str(segment)))
        depth += 1

    for subsegment in segment:
        if isinstance(subsegment, Segment):
            print_tree(subsegment, depth=depth)



from ..util import LookAhead
from .segmenter import Segmenter
from .segments import MatchableSegment, Segment

WHITESPACE = {"whitespace", "break"}
PARAGRAPH_END = {"break"}
SENTENCE_END = {"period", "epoint", "qmark", "tab_open"}
SUB_OPEN = {"brack_open", "dbrack_open", "paren_open", "ref_open",
            "comment_start", "dcurly_open"}
SUB_CLOSE = {"brack_close", "dbrack_close", "paren_close", "ref_close",
             "comment_end", "dcurly_close"}
MIN_SENTENCE = 3


class ParagraphsSentencesAndWhitespace(Segmenter):
    """
    Constructs a paragraphs, sentences and whitespace segmenter.  This segmenter
    is intended to be used in western languages where sentences and paragraphs
    are meaningful segments of text content.

    Tree structure:

    * whitespace : :class:`~deltas.Segment`
    * paragraph : :class:`~deltas.MatchableSegment`
      * sentence : :class:`~deltas.MatchableSegment`
      * whitespace : :class:`~deltas.Segment`

    :Example:
        >>> from deltas import ParagraphsSentencesAndWhitespace, text_split
        >>> from deltas.segmenters import print_tree
        >>>
        >>> a = text_split.tokenize("This comes first.  This comes second.")
        >>>
        >>> segmenter = ParagraphsSentencesAndWhitespace()
        >>> segments = segmenter.segment(a)
        >>>
        >>> print_tree(segments)
        Segment: 'This comes first.  This comes second.'
                MatchableSegment: 'This comes first.  This comes second.'
                        MatchableSegment: 'This comes first.'
                        Segment: '  '
                        MatchableSegment: 'This comes second.'

    :Parameters:
        whitespace : `set` ( `str` )
            A set of token types that represent whitespace.
        paragraph_end : `set` ( `str` )
            A set of token types that represent the end of a pragraph.
        sentence_end : `set` ( `str`)
            A set of tokens types that represent the end of a sentence.
        min_sentence : `int`
            The minimum non-whitespace tokens that a sentence must contain
            before a sentence_end will be entertained.
    """  # noqa

    def __init__(self, *, whitespace=None, paragraph_end=None,
                 sentence_end=None, min_sentence=None):

        self.whitespace = set(whitespace or WHITESPACE)
        self.paragraph_end = set(paragraph_end or PARAGRAPH_END)
        self.sentence_end = set(sentence_end or SENTENCE_END)
        self.min_sentence = int(min_sentence or MIN_SENTENCE)

    def segment(self, tokens):
        """
        Segments a sequence of tokens into a sequence of segments.

        :Parameters:
            tokens : `list` ( :class:`~deltas.Token` )
        """
        look_ahead = LookAhead(tokens)

        segments = Segment()

        while not look_ahead.empty():

            if look_ahead.peek().type not in self.whitespace:  # Paragraph!
                paragraph = MatchableSegment(look_ahead.i)

                while not look_ahead.empty() and \
                        look_ahead.peek().type not in self.paragraph_end:

                    if look_ahead.peek().type == "tab_open":  # Table
                        tab_depth = 1
                        sentence = MatchableSegment(
                            look_ahead.i, [next(look_ahead)])
                        while not look_ahead.empty() and tab_depth > 0:
                            tab_depth += look_ahead.peek().type == "tab_open"
                            tab_depth -= look_ahead.peek().type == "tab_close"
                            sentence.append(next(look_ahead))

                        paragraph.append(sentence)

                    elif look_ahead.peek().type not in self.whitespace:  # Sentence!
                        sentence = MatchableSegment(
                            look_ahead.i, [next(look_ahead)])
                        sub_depth = int(sentence[0].type in SUB_OPEN)
                        while not look_ahead.empty():

                            sub_depth += look_ahead.peek().type in SUB_OPEN
                            sub_depth -= look_ahead.peek().type in SUB_CLOSE
                            sentence.append(next(look_ahead))

                            if sentence[-1].type in \
                                    self.sentence_end and sub_depth <= 0:
                                non_whitespace = sum(
                                    s.type not in self.whitespace
                                    for s in sentence)
                                if non_whitespace >= self.min_sentence:
                                    break

                        paragraph.append(sentence)

                    else:  # look_ahead.peek().type in self.whitespace
                        whitespace = Segment(look_ahead.i, [next(look_ahead)])
                        paragraph.append(whitespace)

                segments.append(paragraph)
            else:  # look_ahead.peek().type in self.whitespace
                whitespace = Segment(look_ahead.i, [next(look_ahead)])
                segments.append(whitespace)

        return segments

    @classmethod
    def from_config(cls, config, name, section_key="segmenters"):
        section = config[section_key][name]
        return cls(
            whitespace=section.get('whitespace'),
            paragraph_end=section.get('paragraph_end'),
            sentence_end=section.get('sentence_end'),
            min_sentence=section.get('min_sentence')
        )



from nose.tools import eq_
import pickle

from ...tokenizers import Token
from ..segments import MatchableSegment, Segment


def test_matchable_segment():

    words = ["foo", "bar", "baz"]

    ms = MatchableSegment(0, [Token(c) for c in enumerate(words)])
    eq_(ms.start, 0)
    eq_(ms.end, len(words))
    hash(ms)
    eq_(ms.match, None)
    ms.match = "derp"
    eq_(ms.match, "derp")

    d = {}
    d[ms] = ms

    ms2 = MatchableSegment(0, [Token(c) for c in enumerate(words)])
    assert ms2 in d


def test_segment():

    words = ["foo", "bar", "baz"]

    ms = Segment(0, [Token(c) for c in enumerate(words)])
    eq_(ms.start, 0)
    eq_(ms.end, len(words))


def test_equality():
    print(hash(MatchableSegment(0, [Token("zero"), Token("one")])))
    print(hash(MatchableSegment(2, [Token("zero"), Token("one")])))
    eq_(MatchableSegment(0, [Token("zero"), Token("one")]),
        MatchableSegment(2, [Token("zero"), Token("one")]))
    eq_(MatchableSegment(0, [
        MatchableSegment(0, [Token("zero"), Token("one")]),
        MatchableSegment(2, [Token("two"), Token("three")])
    ]),
        MatchableSegment(4, [
            MatchableSegment(4, [Token("zero"), Token("one")]),
            MatchableSegment(6, [Token("two"), Token("three")])
        ])
    )


def test_pickling():
    segment = MatchableSegment(0, [
        MatchableSegment(0, [Token("zero"), Token("one")]),
        Segment(2, [Token("two"), Token("three")])
    ])

    unpickled_segment = pickle.loads(pickle.dumps(segment))
    eq_(list(segment.tokens()),
        list(unpickled_segment.tokens()))



"""
Apply operations
================
"""

from .operations import Delete, Equal, Insert


def apply(operations, a_tokens, b_tokens):
    """
    Applies a sequences of operations to tokens -- copies tokens from
    `a_tokens` and `b_tokens` according to `operations`.

    :Parameters:
        operations : sequence of :~class:`deltas.Operation`
            Operations to perform
        a_tokens : list of `comparable`
            Starting sequence of comparable tokens
        b_tokens : list of `comparable`
            Ending list of comparable tokens

    :Returns:
        A new list of tokens
    """
    for operation in operations:

        if isinstance(operation, Equal):
            # print(
            # "Equal: {0}".format(str(a_tokens[operation.a1:operation.a2])))
            for t in a_tokens[operation.a1:operation.a2]:
                yield t

        elif isinstance(operation, Insert):
            # print(
            # "Insert: {0}".format(str(b_tokens[operation.b1:operation.b2])))
            for t in b_tokens[operation.b1:operation.b2]:
                yield t

        elif isinstance(operation, Delete):
            # print(
            # "Delete: {0}".format(str(a_tokens[operation.a1:operation.a2])))
            pass

        else:
            raise TypeError("Unexpected operation type " +
                            "{0}".format(type(operation)))



import re

import yamlconf

from .token import Token

import jieba as ch_jieba
from sudachipy import tokenizer as jp_tokenizer
from sudachipy import dictionary as jp_dictionary
from konlpy.tag import Okt as ko_okt

CH_JIEBA = ch_jieba.Tokenizer()
JAP_SUDACHY = None
KOR_KONLPY_OKT = None


def get_ch_tokenizer():
    if CH_JIEBA.initialized is False:
        CH_JIEBA.initialize()
    return CH_JIEBA


def get_jap_tokenizer():
    global JAP_SUDACHY
    if JAP_SUDACHY is None:
        JAP_SUDACHY = jp_dictionary.Dictionary().create()
    return JAP_SUDACHY


def get_kor_tokenizer():
    global KOR_KONLPY_OKT
    if KOR_KONLPY_OKT is None:
        KOR_KONLPY_OKT = ko_okt()
    return KOR_KONLPY_OKT


class Tokenizer:
    """
    Constructs a tokenizaton strategy.
    """
    def tokenize(self, text, token_class=Token):
        """
        Tokenizes a text.
        """
        raise NotImplementedError()

    @classmethod
    def from_config(cls, config, name, section_key="tokenizers"):
        section = config[section_key][name]
        if 'module' in section:
            return yamlconf.import_module(section['module'])
        else:
            Tokenizer = yamlconf.import_module(section['class'])
            return Tokenizer.from_config(config, name, section_key)


class TokenizerPipeline(Tokenizer):
    def __init__(self, tokenizer, *token_processors):
        self.tokenizer = tokenizer  # start of the pipeline
        self.token_processors = token_processors

    def tokenize(self, text):
        tokens = self.tokenizer.tokenize(text)
        for token_processor in self.token_processors:
            tokens = token_processor.process(tokens)
        return tokens


class RegexTokenizer(Tokenizer):
    """
    Uses a lexicon of regular expressions and names to tokenize a text string.
    """
    def __init__(self, lexicon):
        self.lexicon = lexicon
        self.regex = re.compile('|'.join('(?P<{0}>{1})'.format(name, pattern)
                                         for name, pattern in lexicon))

    def tokenize(self, text, token_class=None):
        return [t for t in self._tokenize(text, token_class=token_class)]

    def _tokenize(self, text, token_class=None):
        """
        Tokenizes a text

        :Returns:
            A `list` of tokens
        """
        token_class = token_class or Token
        tokens = {}

        for i, match in enumerate(self.regex.finditer(text)):
            value = match.group(0)

            try:
                token = tokens[value]
            except KeyError:
                type = match.lastgroup
                token = token_class(value, type=type)
                tokens[value] = token

            yield token


class TokenProcessor:
    def process(self, tokenized_text):
        raise NotImplementedError()


class CJKProcessor(TokenProcessor):
    """
    Uses a cjk_lexicon to decide which tokenizer should be used
    (Chinese, Japanese or Korean).
    """
    def __init__(self, cjk_lexicon, lng_frac_par=0.25):
        self.regex_cjk = re.compile(cjk_lexicon['cjk'])
        self.regex_japanese = re.compile(cjk_lexicon['japanese'])
        self.regex_korean = re.compile(cjk_lexicon['korean'])
        self.lng_frac_par = lng_frac_par

    def process(self, tokenized_text, token_class=None):
        token_class = token_class or Token
        language = self._lng_decision(tokenized_text, self.lng_frac_par)
        processed_tokens = self._cjk_processing(tokenized_text,
                                                language=language,
                                                token_class=token_class)
        return processed_tokens

    def _lng_decision(self, tokenized_text, lng_frac_par=0.25):
        text = "".join([tok[:]
                        for tok in tokenized_text
                        if tok.type == 'cjk_word'])
        cjk_symbols = len(self.regex_cjk.findall(text))
        jap_symbols = len(self.regex_japanese.findall(text))
        kor_symbols = len(self.regex_korean.findall(text))
        char_lng_frac = {'japanese': jap_symbols/cjk_symbols,
                         'korean': kor_symbols/cjk_symbols}
        max_char_lng_frac = max(char_lng_frac, key=char_lng_frac.get)
        # check if at least 1/4 of chars are other than chinese,
        # if not -> run chinese tokenizer
        if char_lng_frac[max_char_lng_frac] > lng_frac_par:
            language = max_char_lng_frac
        else:
            language = 'cjk'
        return language

    def _cjk_processing(self, tokenized_text, language, token_class=None):
        token_class = token_class or Token
        cjk_word_indices = list(filter(
                                lambda x:
                                tokenized_text[x].type == 'cjk_word',
                                range(len(tokenized_text))))

        if language == 'cjk':
            seg = get_ch_tokenizer()
            for i in cjk_word_indices[::-1]:
                processed_cjk_token = seg.lcut(tokenized_text[i],
                                               cut_all=False)
                tokenized_text[i:i+1] = [token_class(word, type="cjk_word")
                                         for word in processed_cjk_token]
            return tokenized_text

        if language == 'japanese':
            mode = jp_tokenizer.Tokenizer.SplitMode.B
            seg = get_jap_tokenizer()
            for i in cjk_word_indices[::-1]:
                processed_cjk_token = [m.surface()
                                       for m in seg.tokenize(str(tokenized_text[i]), # noqa
                                       mode)]
                tokenized_text[i:i+1] = [token_class(word, type="cjk_word")
                                         for word in processed_cjk_token]
            return tokenized_text

        if language == 'korean':
            seg = get_kor_tokenizer()
            for i in cjk_word_indices[::-1]:
                processed_cjk_token = seg.nouns(tokenized_text[i])
                if processed_cjk_token != []:
                    tokenized_text[i:i+1] = [token_class(word, type="cjk_word")
                                             for word in processed_cjk_token]
            return tokenized_text



"""
The delta between sequences of tokens is represented by a sequence of
:class:`~deltas.Operation`.  The goal of difference algorithms is to
detect a sequence of :class:`~deltas.Operation` that has desirable
properties.  :func:`~deltas.apply` can be used with a sequence of
:class:`~deltas.Operation` to convert an initial sequence of tokens
into a changed sequence of tokens.

Specifically, this library understands and produces three types of operations:

* :class:`~deltas.Delete` -- some tokens were deleted
* :class:`~deltas.Insert` -- some tokens were inserted
* :class:`~deltas.Equal` -- some tokens were copied

.. autoclass:: deltas.Delete
.. autoclass:: deltas.Insert
.. autoclass:: deltas.Equal

.. autoclass:: deltas.Operation
"""
from collections import namedtuple

Operation = namedtuple("Operation", ['name', 'a1', 'a2', 'b1', 'b2'])
"""
Represents an option performed on a sequence of tokens to arrive at another
sequence of tokens.  Instances of this type are compatible with the output of
:meth:`difflib.SequenceMatcher.get_opcodes`.
"""

""" This operation is useless and will be ignored
class Replace(Operation):
    def __init__(self, a1, a2, b1, b2):
        Operation.__init__(self, "replace", a1, a2, b1, b2)
"""


class Delete(Operation):
    """
    Represents the deletion of tokens.

    :Parameters:
        a1 : int
            Start position in first sequence.
        a2 : int
            End position in first sequence.
        b1 : int
            Start position in second sequence.
        b2 : int
            End position in second sequence.
    """

    OPNAME = "delete"

    def __new__(cls, a1, a2, b1, b2, name=None):
        return Operation.__new__(cls, "delete", a1, a2, b1, b2)

    def relevant_tokens(self, a, b):
        return a[self.a1:self.a2]


class Insert(Operation):
    """
    Represents the insertions of tokens.

    :Parameters:
        a1 : int
            Start position in first sequence.
        a2 : int
            End position in first sequence.
        b1 : int
            Start position in second sequence.
        b2 : int
            End position in second sequence.
    """

    OPNAME = "insert"

    def __new__(cls, a1, a2, b1, b2, name=None):
        return Operation.__new__(cls, "insert", a1, a2, b1, b2)

    def relevant_tokens(self, a, b):
        return b[self.b1:self.b2]


class Equal(Operation):
    """
    Represents the equality of tokens between sequences.

    :Parameters:
        a1 : int
            Start position in first sequence.
        a2 : int
            End position in first sequence.
        b1 : int
            Start position in second sequence.
        b2 : int
            End position in second sequence.
    """

    OPNAME = "equal"

    def __new__(cls, a1, a2, b1, b2, name=None):
        return Operation.__new__(cls, "equal", a1, a2, b1, b2)

    def relevant_tokens(self, a, b):
        return a[self.a1:self.a2]


def print_operations(operations, a, b):
    for operation in operations:
        print("{0}: '{1}'".format(operation.name,
                                  ''.join(operation.relevant_tokens(a, b))))



"""
Performs a diffs using a tree of matchable segments in order to remain robust
to content moves.  This module supports the use of a custom
:class:`~deltas.Segmenter`.

.. autofunction:: deltas.algorithms.segment_matcher.diff

.. autofunction:: deltas.algorithms.segment_matcher.diff_segments

.. autofunction:: deltas.algorithms.segment_matcher.process

.. autoclass:: deltas.SegmentMatcher
    :members:
"""
from collections import defaultdict

from . import sequence_matcher
from ..operations import Delete, Equal, Insert
from ..segmenters import (MatchableSegment, ParagraphsSentencesAndWhitespace,
                          Segment, Segmenter)
from ..tokenizers import Token, Tokenizer, text_split
from .diff_engine import DiffEngine

SEGMENTER = ParagraphsSentencesAndWhitespace()
TOKENIZER = text_split


def diff(a, b, segmenter=None):
    """
    Performs a diff comparison between two sequences of tokens (`a` and `b`)
    using `segmenter` to cluster and match
    :class:`deltas.MatchableSegment`.

    :Example:
        >>> from deltas import segment_matcher, text_split
        >>>
        >>> a = text_split.tokenize("This is some text.  This is some other text.")
        >>> b = text_split.tokenize("This is some other text.  This is some text.")
        >>> operations = segment_matcher.diff(a, b)
        >>>
        >>> for op in operations:
        ...     print(op.name, repr(''.join(a[op.a1:op.a2])),
        ...           repr(''.join(b[op.b1:op.b2])))
        ...
        equal 'This is some other text.' 'This is some other text.'
        insert '' '  '
        equal 'This is some text.' 'This is some text.'
        delete '  ' ''

    :Parameters:
        a : `list`(:class:`deltas.tokenizers.Token`)
            Initial sequence
        b : `list`(:class:`deltas.tokenizers.Token`)
            Changed sequence
        segmenter : :class:`deltas.Segmenter`
            A segmenter to use on the tokens.

    :Returns:
        An `iterable` of operations.
    """
    a, b = list(a), list(b)
    segmenter = segmenter or SEGMENTER

    # Cluster the input tokens
    a_segments = segmenter.segment(a)
    b_segments = segmenter.segment(b)

    return diff_segments(a_segments, b_segments)


def diff_segments(a_segments, b_segments):
    """
    Performs a diff comparison between two pre-clustered
    :class:`deltas.Segment` trees.  In most cases, segmentation
    takes 100X more time than actually performing the diff.

    :Parameters:
        a_segments : :class:`deltas.Segment`
            An initial sequence
        b_segments : :class:`deltas.Segment`
            A changed sequence

    :Returns:
        An `iterable` of operations.
    """
    # Match and re-sequence unmatched tokens
    a_segment_tokens, b_segment_tokens = _cluster_matching_segments(a_segments,
                                                                    b_segments)

    # Perform a simple LCS over unmatched tokens and clusters
    clustered_ops = sequence_matcher.diff(a_segment_tokens, b_segment_tokens)

    # Return the expanded (de-clustered) operations
    return (op for op in SegmentOperationsExpander(clustered_ops,
                                                   a_segment_tokens,
                                                   b_segment_tokens).expand())


def process(texts, *args, **kwargs):
    """
    Processes a single sequence of texts with a
    :class:`~deltas.SegmentMatcher`.

    :Parameters:
        texts : `iterable`(`str`)
            sequence of texts
        args : `tuple`
            passed to :class:`~deltas.SegmentMatcher`'s
            constructor
        kwaths : `dict`
            passed to :class:`~deltas.SegmentMatcher`'s
            constructor
    """
    processor = SegmentMatcher.Processor(*args, **kwargs)
    for text in texts:
        yield processor.process(text)


class SegmentMatcher(DiffEngine):
    """
    Constructs a segment matcher diff engine that preserves segmentation state
    and is able to process changes sequentially.  When detecting changes
    across many versions of a text this strategy will be about twice as fast as
    calling :func:`diff` sequentially.

    :Example:
        >>> from deltas import SegmentMatcher
        >>> from deltas import text_split
        >>>
        >>> engine = SegmentMatcher(text_split)
        >>>
        >>> processor = engine.processor()
        >>> ops, a, b = processor.process("This is a version.  It has some " +
        ...                               "text in it.")
        >>> print(" ".join(repr(''.join(b[op.b1:op.b2])) for op in ops))
        'This is a version.  It has some text in it.'
        >>> ops, a, b = processor.process("This is a version.  However, it " +
        ...                               "has different.")
        >>> print(" ".join(repr(''.join(b[op.b1:op.b2])) for op in ops))
        'This is a version.  ' '' 'However, it' ' has ' '' 'different' '.'
        >>> ops, a, b = processor.process("Switching it up here.  This is a " +
        ...                               "version.")
        >>> print(" ".join(repr(''.join(b[op.b1:op.b2])) for op in ops))
        '' 'Switching' ' it ' '' 'up' ' ' '' 'here' '.' '  ' 'This is a version.'
    """  # noqa

    class Processor(DiffEngine.Processor):
        """
        A processor used by the SegmentMatcher difference engine to track the
        history of a single text.
        """

        def __init__(self, tokenizer=None, segmenter=None, last_text=None,
                     last_tokens=None, last_segments=None):
            self.tokenizer = tokenizer or TOKENIZER
            self.segmenter = segmenter or SEGMENTER
            self.update(last_text, last_tokens, last_segments)

        def update(self, last_text=None, last_tokens=None, last_segments=None):
            if last_segments is not None:
                self.last_segments = last_segments
                self.last_tokens = self.last_segments.tokens()
            elif last_tokens is not None:
                self.last_tokens = last_tokens
                self.last_segments = self.segmenter.segment(last_tokens)
            elif last_text is not None:
                self.last_tokens = self.tokenizer.tokenize(last_text)
                self.last_segments = self.segmenter.segment(self.last_tokens)
            else:
                self.last_tokens = []
                self.last_segments = Segment()

        def process(self, text, token_class=Token):
            """
            Processes a new version of a text and returns the delta.

            :Parameters:
                text : `str`
                    The text to process

            :Returns:
                    A tuple of `operations`, `a_tokens`, `b_tokens`
            """
            # Tokenize and segment
            tokens = self.tokenizer.tokenize(text, token_class=token_class)
            segments = self.segmenter.segment(tokens)

            return self.process_segments(segments, tokens=tokens)

        def process_segments(self, segments, tokens=None):

            if tokens is None:
                tokens = segments.tokens()

            # Perform diff
            _clear_matches(self.last_segments)
            operations = diff_segments(self.last_segments, segments)

            # Update state
            a = self.last_tokens
            b = tokens
            self.last_tokens = tokens
            self.last_segments = segments

            # Return delta
            return operations, a, b

    def __init__(self, tokenizer=None, segmenter=None):
        self.tokenizer = tokenizer or TOKENIZER
        self.segmenter = segmenter or SEGMENTER

    def processor(self, *args, **kwargs):
        """
        Constructs and configures a processor to process versions of a text.
        """
        return self.Processor(self.tokenizer, self.segmenter, *args, **kwargs)

    def process(self, texts, *args, **kwargs):
        return process(texts, self.tokenizer, self.segmenter, *args, **kwargs)

    @classmethod
    def from_config(cls, config, name, section_key="diff_engines"):
        section = config[section_key][name]
        return cls(
            Tokenizer.from_config(config, section['tokenizer']),
            Segmenter.from_config(config, section['segmenter'])
        )


def _cluster_matching_segments(a_segments, b_segments):

    # Generate a look-up map for matchable segments in 'a'
    a_segment_map = _build_segment_map(a_segments)

    # Find and cluster matching content in 'b'
    b_segment_tokens = list(_match_segments(a_segment_map, b_segments))

    # Expand unmatched segments from 'a'
    a_segment_tokens = list(_expand_unmatched_segments(a_segments))

    return a_segment_tokens, b_segment_tokens


def _build_segment_map(segments):
    d = defaultdict(list)
    for matchable_segment in _get_matchable_segments(segments):
        d[matchable_segment].append(matchable_segment)

    return d


def _get_matchable_segments(segments):
    """
    Performs a depth-first search of the segment tree to get all matchable
    segments.
    """
    for subsegment in segments:
        if isinstance(subsegment, Token):
            break  # No tokens allowed next to segments
        if isinstance(subsegment, Segment):
            if isinstance(subsegment, MatchableSegment):
                yield subsegment

            for matchable_subsegment in _get_matchable_segments(subsegment):
                yield matchable_subsegment


def _match_segments(a_segment_map, b_segments):
    for subsegment in b_segments:
        if isinstance(subsegment, Segment):

            if isinstance(subsegment, MatchableSegment) and \
               subsegment in a_segment_map:
                matched_segments = a_segment_map[subsegment]  # Get matches
                for matched_segment in matched_segments:  # For each match
                    matched_segment.match = subsegment  # flag as matched
                subsegment.match = matched_segments[0]  # first match
                yield subsegment  # Dump matched segment

            else:
                for seg_or_tok in _match_segments(a_segment_map, subsegment):
                    yield seg_or_tok  # Recurse

        else:
            yield subsegment  # Dump token


def _expand_unmatched_segments(a_segments):
    for subsegment in a_segments:
        # Check if a segment is matched.
        if isinstance(subsegment, Segment):

            if isinstance(subsegment, MatchableSegment) and \
               subsegment.match is not None:
                yield subsegment  # Yield matched segment as cluster
            else:
                for seg_or_tok in _expand_unmatched_segments(subsegment):
                    yield seg_or_tok  # Recurse
        else:
            yield subsegment  # Dump token


def _clear_matches(segment):
    if isinstance(segment, MatchableSegment):
        segment.match = None

    if isinstance(segment, Segment):
        # Recurse!
        for subsegment in segment:
            _clear_matches(subsegment)


class SegmentOperationsExpander:

    def __init__(self, operations, a_token_segments, b_token_segments):

        self.a_pos = 0
        self.b_pos = 0
        self.a_token_segments = a_token_segments
        self.b_token_segments = b_token_segments
        self.operations = operations

    def expand(self):
        for operation in self.operations:
            if isinstance(operation, Equal):
                # print(
                # "Processing equal: {0} {1}".format(self.a_pos, self.b_pos))
                expanded_operations = self._process_equal(operation)
            elif isinstance(operation, Insert):
                # print(
                # "Processing insert: {0} {1}".format(self.a_pos, self.b_pos))
                expanded_operations = self._process_insert(operation)
            elif isinstance(operation, Delete):
                # print(
                # "Processing remove: {0} {1}".format(self.a_pos, self.b_pos))
                expanded_operations = self._process_delete(operation)
            else:
                assert False, "Should never happen"

            for operation in expanded_operations:
                yield operation

    def _process_equal(self, op):
        a1 = self.a_pos
        b1 = self.b_pos
        token_len = sum(1 for t_s in self.a_token_segments[op.a1:op.a2]
                        for _ in t_s.tokens())
        self.a_pos += token_len
        self.b_pos += token_len

        yield Equal(a1, self.a_pos, b1, self.b_pos)

    def _process_insert(self, op):
        inserted_token_count = 0

        for t_s in self.b_token_segments[op.b1:op.b2]:
            if isinstance(t_s, Token):
                inserted_token_count += 1
            else:  # Found a matched segment
                segment = t_s

                # First, emit an insert for the tokens we have seen so far
                if inserted_token_count > 0:
                    b1 = self.b_pos
                    self.b_pos += inserted_token_count
                    yield Insert(self.a_pos, self.a_pos, b1, self.b_pos)
                    inserted_token_count = 0

                # Now, emit an Equal for the matched segment
                b1 = self.b_pos
                self.b_pos += sum(1 for _ in segment.tokens())
                yield Equal(segment.match.start, segment.match.end,
                            b1, self.b_pos)

        # Cleanup!  Make sure we emit any remaining inserted tokens.
        if inserted_token_count > 0:
            b1 = self.b_pos
            self.b_pos += inserted_token_count
            yield Insert(self.a_pos, self.a_pos, b1, self.b_pos)
            inserted_token_count = 0

    def _process_delete(self, op):
        removed_token_count = 0
        for t_s in self.a_token_segments[op.a1:op.a2]:

            if isinstance(t_s, Token):
                removed_token_count += 1
            else:  # Found a matched token... not removed -- just moved
                segment = t_s

                if removed_token_count > 0:
                    a1 = self.a_pos
                    self.a_pos += removed_token_count
                    yield Delete(a1, self.a_pos, self.b_pos, self.b_pos)
                    removed_token_count = 0

                # update & reset!
                self.a_pos += sum(1 for _ in segment.tokens())

        # cleanup
        if removed_token_count > 0:
            a1 = self.a_pos
            self.a_pos += removed_token_count
            yield Delete(a1, self.a_pos, self.b_pos, self.b_pos)



"""
Text segmentation is the process of dividing written text into meaningful units,
such as words, sentences, or topics.  This module provides a collection of
:class:`~deltas.Segmenter` that can be used to produce hierarchical
clusters of tokens (:class:`~deltas.Segmenter`) that can be
understood by :class:`~deltas.algorithms.segment_matcher`.

:class:`~deltas.Segmenter`
    is an abstract base class that requires the implementation of a
    :func:`~deltas.Segmenter.segment`
    function that clusters tokens into a sequences of
    :class:`~deltas.Segment` and
    :class:`~deltas.MatchableSegment`

:class:`~deltas.ParagraphsSentencesAndWhitespace`
    implements a
    :func:`~deltas.ParagraphsSentencesAndWhitespace.segment`
    function that clusters tokens into segments of paragraph and
    sentence :class:`~deltas.MatchableSegment` with whitespace
    :class:`~deltas.Segment` inbetween.
"""

from .paragraphs_sentences_and_whitespace import ParagraphsSentencesAndWhitespace
from .segmenter import Segmenter
from .segments import Segment, MatchableSegment
from .functions import print_tree


__all__ = [ParagraphsSentencesAndWhitespace, Segmenter,
           Segment, MatchableSegment, print_tree]



"""
Tokens represent chuncks of text that have semantic meaning.  A Token class that
extends :class:`str` is provided.

.. autoclass:: deltas.Token
    :members:
"""


class Token(str):
    """
    Constructs a typed sub-string extracted from a text.
    """
    __slots__ = ("type")

    def __new__(cls, content, *args, **kwargs):
        if isinstance(content, cls):
            return content
        else:
            return super().__new__(cls, content)

    def tokens(self):
        """
        Returns an iterator of *self*.  This method reflects the behavior of
        :meth:`deltas.Segment.tokens`
        """
        yield self

    def __init__(self, content, type=None):
        self.type = str(type) if type is not None else None
        """
        An optional value describing the type of token.
        """

    def __repr__(self):
        return "{0}({1}, type={2})" \
               .format(self.__class__.__name__,
                       super().__repr__(),
                       repr(self.type))



from nose.tools import eq_

from ..util import LookAhead


def test_lookahead_list():

    test_list = [1, 2, 3]

    look_ahead = LookAhead(test_list)
    eq_(look_ahead.peek(), 1)
    eq_(look_ahead.pop(), 1)
    eq_(look_ahead.peek(), 2)

    for i, val in enumerate(look_ahead):
        eq_(val, i+2)


def test_lookahead_generator():

    test_list = (i for i in [1, 2, 3])

    look_ahead = LookAhead(test_list)
    eq_(look_ahead.peek(), 1)
    eq_(look_ahead.pop(), 1)
    eq_(look_ahead.peek(), 2)

    for i, val in enumerate(look_ahead):
        eq_(val, i+2)


def test_lookahead_identity():

    l1 = LookAhead([])
    l2 = LookAhead(l1)

    eq_(l1, l2)
    eq_(id(l1), id(l2))  # Confirms same memory location



from nose.tools import eq_

from ..apply import apply


def diff_sequence(process, tokenizer=None):
    revisions = ["This sentence is a sentence.  This is gonna go.",
                 "This sentence is a sentence.  Hi!  I'm new here.",
                 "Hi!  I'm new here.  This sentence is a sentence. ",
                 "I'm new here.  This sentence is a sentence. I'm new here.",
                 "I'm new here. Sentence is a sentence."]

    token_operations = process(revisions)

    tokens = []
    for ops, a, b in token_operations:
        tokens = apply(ops, a, b)
        print("|".join(str(t) for t in b))

    eq_(''.join(tokens), revisions[-1])



from nose.tools import eq_

from ...tokenizers import wikitext_split
from ..paragraphs_sentences_and_whitespace import \
    ParagraphsSentencesAndWhitespace
from ..segments import MatchableSegment, Segment


def test_segment():

    segmenter = ParagraphsSentencesAndWhitespace()

    text = 'This is some text.  This is some other text.\n ' + \
           'A. Peterson is a name that <!-- This is derp. --> I made up.\n' + \
           'This is an additional sentence.\n' + \
           '\n' + \
           '== OMG HEADER ==\n' + \
           'This is a new paragraph! Isn\'t this fun?'

    tokens = wikitext_split.tokenize(text)
    segments = list(segmenter.segment(tokens))
    print([str(s) for s in segments[0]])

    eq_(len(segments), 3)  # 2 paragraphs + 1 whitespace

    eq_(len(segments[0]), 8)  # 4 sentences + 4 whitespace

    eq_(type(segments[1]), Segment)
    eq_(type(segments[2]), MatchableSegment)
    eq_(type(segments[2][2]), MatchableSegment)

    eq_(''.join(str(s) for s in segments), text)



"""
Tokenizers are used to split text content into a sequence of tokens.  Extend
:class:`~deltas.Tokenizer` to implement a custom tokenizer.  See
also :class:`~deltas.RegexTokenizer` for producing a tokenizer
based on a lexicon.

**deltas.text_split**
    a :class:`~deltas.RegexTokenizer` that splits text into words,
    punctuation, symbols and whitespace.

**deltas.wikitext_split**
    a :class:`~deltas.RegexTokenizer` that splits text into words,
    punctuation, symbols and whitespace as well as wikitext markup elements
    (e.g. ('dcurly_open', "{{") and ('bold', "'''"))
"""
from .tokenizer import Tokenizer, RegexTokenizer, CJKProcessor, TokenizerPipeline # noqa
from .token import Token
from .text_split import text_split
from .wikitext_split import wikitext_split
from .wikitext_split_w_cjk import wikitext_split_w_cjk


__all__ = [Tokenizer, RegexTokenizer, CJKProcessor, TokenizerPipeline, Token, text_split, wikitext_split, wikitext_split_w_cjk] # noqa



# Unified Ideographs, Unified Ideographs Ext A,
# Unified Ideographs Ext. B, Compatibility Ideographs,
# Compatibility Ideographs Suppl.
chinese_n_misc_char = r'\u4E00-\u62FF' + \
                      r'\u3400-\u4DFF' + \
                      r'\U00020000-\U000215FF' + \
                      r'\uF900-\uFAFF' + \
                      r'\U0002F800-\U0002FA1F' + \
                      r'\u6300-\u77FF' + \
                      r'\u7800-\u8CFF' + \
                      r'\u8D00-\u9FCC' + \
                      r'\U00021600-\U000230FF' + \
                      r'\U00023100-\U000245FF' + \
                      r'\U00024600-\U000260FF' + \
                      r'\U00026100-\U000275FF' + \
                      r'\U00027600-\U000290FF' + \
                      r'\U00029100-\U0002A6DF' + \
                      r'\u4E00-\u9FCB' + \
                      r'\uF900-\uFA6A' + \
                      r'\u3220-\u3243' + \
                      r'\u3280-\u337F'

# Hiragana, Katakana, Kanji, Kanji radicals,
# Katakana and Punctuation (Half Width), Misc Symbols and Chars
jap_char = r'\u3041-\u3096' + \
           r'\u30A0-\u30FF' + \
           r'\u3400-\u4DB5' + \
           r'\u2E80-\u2FD5' + \
           r'\uFF5F-\uFF9F' + \
           r'\u31F0-\u31FF'

# hangul syllables, hangul jamo, hangul, hangul, hangul
kor_char = r'\uAC00-\uD7AF' + \
           r'\u1100-\u11FF' + \
           r'\u3130–\u318F' + \
           r'\uA960–\uA97F' + \
           r'\uD7B0–\uD7FF'

cjk = chinese_n_misc_char + jap_char + kor_char

devangari_word = r'\u0901-\u0963'

arabic_word = r'\u0601-\u061A' + \
              r'\u061C-\u0669' + \
              r'\u06D5-\u06EF'

bengali_word = r'\u0980-\u09FF'



"""
Sequence Matcher
----------------

Performs a simple *longest-common-substring* diff.  This module implements a
simple wrapper around :class:`difflib.SequenceMatcher`.

.. autofunction:: deltas.algorithms.sequence_matcher.diff

.. autofunction:: deltas.algorithms.sequence_matcher.process

.. autoclass:: deltas.SequenceMatcher
    :members:
"""

from difflib import SequenceMatcher as SM

from ..operations import Delete, Equal, Insert
from ..tokenizers import text_split
from .diff_engine import DiffEngine

TOKENIZER = text_split

OP_PARSERS = {
    "replace": lambda a1, a2, b1, b2: [Delete(a1, a2, b1, b1),
                                       Insert(a2, a2, b1, b2)],
    "insert": lambda a1, a2, b1, b2: [Insert(a1, a2, b1, b2)],
    "delete": lambda a1, a2, b1, b2: [Delete(a1, a2, b1, b2)],
    "equal": lambda a1, a2, b1, b2: [Equal(a1, a2, b1, b2)]
}


def diff(a, b):
    """
    Performs a longest common substring diff.

    :Parameters:
        a : sequence of `comparable`
            Initial sequence
        b : sequence of `comparable`
            Changed sequence

    :Returns:
        An `iterable` of operations.
    """
    a, b = list(a), list(b)
    opcodes = SM(None, a, b).get_opcodes()
    return parse_opcodes(opcodes)


def process(texts, *args, **kwargs):
    processor = SequenceMatcher.Processor(*args, **kwargs)

    for text in texts:
        yield processor.process(text)


class SequenceMatcher(DiffEngine):
    """
    Constructs a sequence matching diff engine that preserves verion state
    and is able to process changes sequentially.  When detecting changes
    across many versions of a text this strategy will be about twice as fast as
    calling :func:`diff` sequentially.

    :Example:
        >>> from deltas.algorithms import SequenceMatcher
        >>> engine = SequenceMatcher()
        >>>
        >>> processor = engine.processor()
        >>> ops, a, b = processor.process("This is a version.  It has some " +
        ...                               "text in it.")
        >>> print(" ".join(repr(''.join(b[op.b1:op.b2])) for op in ops))
        'This is a version.  It has some text in it.'
        >>> ops, a, b = processor.process("This is a version.  However, it " +
        ...                               "has different.")
        >>> print(" ".join(repr(''.join(b[op.b1:op.b2])) for op in ops))
        'This is a version.  ' '' 'However, it' ' has ' '' 'different' '.'
        >>> ops, a, b = processor.process("Switching it up here.  This is " +
        ...                               "a version.")
        >>> print(" ".join(repr(''.join(b[op.b1:op.b2])) for op in ops))
        'Switching it up here.  ' 'This is a version.' ''
    """

    class Processor(DiffEngine.Processor):
        __slots__ = ('last_tokens')
        """
        A processor used by the SequenceMatcher difference engine to track the
        history of a single text.
        """
        def __init__(self, tokenizer=None, last_text=None, last_tokens=None,
                     token_class=None):
            self.tokenizer = tokenizer or TOKENIZER
            self.update(last_text, last_tokens)
            self.token_class = token_class

        def update(self, last_text=None, last_tokens=None, **kwargs):
            if last_text is not None:
                self.last_tokens = self.tokenizer.tokenize(last_text, **kwargs)
            else:
                self.last_tokens = last_tokens or []

        def process(self, text, token_class=None):
            """
            Processes a new version of a text and returns the delta.

            :Parameters:
                text : `str`
                    The text to process

            :Returns:
                A tuple of `operations`, `a_tokens`, `b_tokens`
            """
            token_class = token_class or self.token_class
            tokens = self.tokenizer.tokenize(text, token_class=token_class)
            operations = diff(self.last_tokens, tokens)

            a = self.last_tokens
            b = tokens
            self.last_tokens = tokens

            return operations, a, b

    def __init__(self, tokenizer=None):
        self.tokenizer = tokenizer or TOKENIZER

    def processor(self, *args, **kwargs):
        return self.Processor(self.tokenizer, *args, **kwargs)

    def process(self, texts, *args, **kwargs):
        return process(texts, self.tokenizer, *args, **kwargs)

    @classmethod
    def from_config(cls, config, name, section_key="diff_engines"):
        return cls()


def parse_opcodes(opcodes):

    for opcode in opcodes:
        op, a_start, a_end, b_start, b_end = opcode

        parse = OP_PARSERS[op]
        for operation in parse(a_start, a_end, b_start, b_end):
            yield operation



import yamlconf


class Segmenter:
    """
    Constructs a token segmentation strategy.
    """
    def __init__(self):
        pass

    def segment(self, tokens):
        """
        Segments a sequence of :class:`~deltas.Token` into a
        `iterable` of :class:`~deltas.Segment`
        """
        raise NotImplementedError()

    @classmethod
    def from_config(cls, config, name, section_key="segmenters"):
        """
        Constructs a segmenter from a configuration doc.
        """
        section = config[section_key][name]
        segmenter_class_path = section['class']
        Segmenter = yamlconf.import_module(segmenter_class_path)
        return Segmenter.from_config(config, name, section_key=section_key)



from nose.tools import eq_

from ..diff_engine import DiffEngine


def test_from_config():
    doc = {
        'diff_engines': {
            'segment_matcher': {
                'class': "deltas.algorithms.SegmentMatcher",
                'segmenter': "western_psw",
                'tokenizer': "wikitext_split"
            }
        },
        'segmenters': {
            'western_psw': {
                'class': "deltas.segmenters.ParagraphsSentencesAndWhitespace"
            }
        },
        'tokenizers': {
            'wikitext_split': {
                'module': "deltas.wikitext_split"
            }
        }
    }

    segment_matcher = DiffEngine.from_config(doc, "segment_matcher")
    text_operations = list(segment_matcher.process(["Foo bar.", "Foo burp."]))
    eq_(len(text_operations), 2)

    ops, a, b = text_operations[1]
    eq_(len(list(ops)), 4)



"""
Segments represent subsequences of tokens that have interesting properties.  All
segments are based on two abstract types:

:class:`deltas.Segment`
    A segment of text with a ``start`` and ``end`` index that refers to the
    original sequence of tokens.
:class:`deltas.MatchableSegment`
    A segment of text that can be matched with another segment no matter where
    it appears in a document.  Generally segmnents of this type represent a
    substantial collection of tokens.

Segment Types
^^^^^^^^^^^^^

.. autoclass:: deltas.Segment
    :members:

.. autoclass:: deltas.MatchableSegment
    :members:
"""
import hashlib


class Segment(list):
    __slots__ = ("start", )
    """
    Represents a sequence of of tokens.  Note that plain Segments are not
    matchable.  Plain segments are generally reserved for whitespace.  For
    matchable segments, see :class:`~deltas.MatchableSegment`.

    Note that :class:`~deltas.Segment` behaves like a list, but it
    will expect that everything added will be of type
    :class:`~deltas.Segment` or :class:`~deltas.Token`.
    """
    def __new__(cls, *args, **kwargs):
        if len(args) == 1 and len(kwargs) == 0 and isinstance(args[0], cls):
            args[0]
        else:
            inst = super().__new__(cls, *args, **kwargs)
            inst.initialize(*args, **kwargs)
            return inst

    def __init__(self, *args, **kwargs): pass

    def initialize(self, start=0, subsegments=None):
        subsegments = subsegments or []
        super().__init__(subsegments)
        self.start = int(start)

    def tokens(self):
        """
        `generator` : the tokens in this segment
        """
        for subsegment_or_token in self:
            if isinstance(subsegment_or_token, Segment):
                subsegment = subsegment_or_token
                for token in subsegment.tokens():
                    yield token
            else:
                token = subsegment_or_token
                yield token

    @property
    def end(self):
        """
        The index of the last :class:`deltas.Token` in the segment.
        """
        return self.start + sum(1 for _ in self.tokens())

    def __repr__(self):
        return "{0}({1})".format(self.__class__.__name__, super().__repr__())

    def __str__(self):
        return ''.join(str(ss) for ss in self)

    def __eq__(self, other):
        raise NotImplementedError()

    def __neq__(self, other):
        raise NotImplementedError()

    def __hash__(self, other):
        raise NotImplementedError()


class MatchableSegment(Segment):
    """
    Constructs a segment that can be matched.  Segments of this type general
    contain important content that might have been copied between different
    versions of text.
    """
    __slots__ = ("sha1", "match")

    def initialize(self, *args, **kwargs):
        super().initialize(*args, **kwargs)
        self.sha1 = hashlib.sha1(bytes(str(self), 'utf-8'))
        self.match = None

    def __eq__(self, other):
        try:
            return hash(self) == hash(other)
        except AttributeError:
            return False

    def __neq__(self, other):
        try:
            return hash(self) != hash(other)
        except AttributeError:
            return False

    def __hash__(self):
        return hash(self.sha1.digest())

    def __getstate__(self): return self.start, list(self)

    def __setstate__(self, args): self.initialize(*args)

    def append(self, subsegment):
        super().append(subsegment)
        self.sha1.update(bytes(str(subsegment), 'utf-8'))

    def extend(self, subsegments):
        for subsegment in subsegments:
            self.append(subsegment)



from nose.tools import eq_

from ..wikitext_split import wikitext_split


def test_wikitext_split():

    input = "As a sentence, this 34 includes punctuation. \n" + \
            "\n" + \
            "==Header!==\n" + \
            "मादरचोद मादरचोद " + \
            "مُنیر " + \
            "克·科伊尔 し〤。foobar!" + \
            "And then we have another sentence here!\n" + \
            "[//google.com foo] " + \
            "https://website.gov?param=value\n" + \
            "peoples' ain't d’encyclopédie\n" + \
            "<ref>derp</ref><ref name=\"foo\" />" + \
            "[[foo|bar]]" + \
            "mailto:email@email.mail " + \
            "위키백과의 운영은 비영리 단체인 위키미디어 재단이 " + \
            "দেখার পর তিনি চ্চিত্র " + \
            "'''some bold''' text m80" + \
            "<ref name=\"doi10.1093/qjmam/1.1.287\">" + \
            "https://web.archive.org/web/20150905180420/" \
            "http://www.turing.org.uk/sources/biblio3.html|archive-date=5"

    expected = [('As', 'word'),
                (' ', 'whitespace'),
                ('a', 'word'),
                (' ', 'whitespace'),
                ('sentence', 'word'),
                (',', 'comma'),
                (' ', 'whitespace'),
                ('this', 'word'),
                (' ', 'whitespace'),
                ('34', 'number'),
                (' ', 'whitespace'),
                ('includes', 'word'),
                (' ', 'whitespace'),
                ('punctuation', 'word'),
                ('.', 'period'),
                (' ', 'whitespace'),
                ('\n\n', 'break'),
                ('==', 'equals'),
                ('Header', 'word'),
                ('!', 'epoint'),
                ('==', 'equals'),
                ('\n', 'whitespace'),
                ('मादरचोद', 'word'),
                (' ', 'whitespace'),
                ('मादरचोद', 'word'),
                (' ', 'whitespace'),
                ('مُنیر', 'word'),
                (' ', 'whitespace'),
                ('克', 'cjk_word'),
                ('·', 'etc'),
                ('科伊尔', 'cjk_word'),
                (' ', 'whitespace'),
                ('し', 'cjk_word'),
                ('〤', 'japan_punct'),
                ('。', 'japan_punct'),
                ('foobar', 'word'),
                ('!', 'epoint'),
                ('And', 'word'),
                (' ', 'whitespace'),
                ('then', 'word'),
                (' ', 'whitespace'),
                ('we', 'word'),
                (' ', 'whitespace'),
                ('have', 'word'),
                (' ', 'whitespace'),
                ('another', 'word'),
                (' ', 'whitespace'),
                ('sentence', 'word'),
                (' ', 'whitespace'),
                ('here', 'word'),
                ('!', 'epoint'),
                ('\n', 'whitespace'),
                ('[', 'brack_open'),
                ('//google.com', 'url'),
                (' ', 'whitespace'),
                ('foo', 'word'),
                (']', 'brack_close'),
                (' ', 'whitespace'),
                ('https://website.gov?param=value', 'url'),
                ('\n', 'whitespace'),
                ("peoples'", 'word'),
                (' ', 'whitespace'),
                ("ain't", 'word'),
                (' ', 'whitespace'),
                ('d’encyclopédie', 'word'),
                ('\n', 'whitespace'),
                ('<ref>', 'ref_open'),
                ('derp', 'word'),
                ('</ref>', 'ref_close'),
                ('<ref name="foo" />', 'ref_singleton'),
                ('[[', 'dbrack_open'),
                ('foo', 'word'),
                ('|', 'bar'),
                ('bar', 'word'),
                (']]', 'dbrack_close'),
                ('mailto:email@email.mail', 'url'),
                (' ', 'whitespace'),
                ('위키백과의', 'cjk_word'),
                (' ', 'whitespace'),
                ('운영은', 'cjk_word'),
                (' ', 'whitespace'),
                ('비영리', 'cjk_word'),
                (' ', 'whitespace'),
                ('단체인', 'cjk_word'),
                (' ', 'whitespace'),
                ('위키미디어', 'cjk_word'),
                (' ', 'whitespace'),
                ('재단이', 'cjk_word'),
                (' ', 'whitespace'),
                ('দেখার', 'word'),
                (' ', 'whitespace'),
                ('পর', 'word'),
                (' ', 'whitespace'),
                ('তিনি', 'word'),
                (' ', 'whitespace'),
                ('চ্চিত্র', 'word'),
                (' ', 'whitespace'),
                ("'''", 'bold'),
                ('some', 'word'),
                (' ', 'whitespace'),
                ('bold', 'word'),
                ("'''", 'bold'),
                (' ', 'whitespace'),
                ('text', 'word'),
                (' ', 'whitespace'),
                ('m80', 'word'),
                ('<ref name="doi10.1093/qjmam/1.1.287">', 'ref_open'),
                ('https://web.archive.org/web/20150905180420/' +
                 'http://www.turing.org.uk/sources/biblio3.html', 'url'),
                ('|', 'bar'),
                ('archive', 'word'),
                ('-', 'etc'),
                ('date', 'word'),
                ('=', 'equals'),
                ('5', 'number')]

    tokens = list(wikitext_split.tokenize(input))

    for token, (s, t) in zip(tokens, expected):
        print(repr(token), (s, t))
        eq_(token, s)
        eq_(token.type, t)


def test_arabic():
    input = "يرجع الأمويون في نسبهم إلى أميَّة بن عبد شمس من قبيلة قريش."
    expected = [("يرجع", "word"),
                (" ", "whitespace"),
                ("الأمويون", "word"),
                (" ", "whitespace"),
                ("في", "word"),
                (" ", "whitespace"),
                ("نسبهم", "word"),
                (" ", "whitespace"),
                ("إلى", "word"),
                (" ", "whitespace"),
                ("أميَّة", "word"),
                (" ", "whitespace"),
                ("بن", "word"),
                (" ", "whitespace"),
                ("عبد", "word"),
                (" ", "whitespace"),
                ("شمس", "word"),
                (" ", "whitespace"),
                ("من", "word"),
                (" ", "whitespace"),
                ("قبيلة", "word"),
                (" ", "whitespace"),
                ("قريش", "word"),
                (".", "period")]

    tokens = list(wikitext_split.tokenize(input))

    for token, (s, t) in zip(tokens, expected):
        print(repr(token), (s, t))
        eq_(token, s)
        eq_(token.type, t)


def test_hebrew():
    input = 'דגל קנדה הידוע בכינויו "דגל עלה האדר" (או המייפל) אומץ בשנת ' + \
            '1965 לאחר דיון ציבורי סביב שאלת הסמלים הלאומיים שבו.'

    expected = [("דגל", "word"),
                (" ", "whitespace"),
                ("קנדה", "word"),
                (" ", "whitespace"),
                ("הידוע", "word"),
                (" ", "whitespace"),
                ("בכינויו", "word"),
                (" ", "whitespace"),
                ('"', "etc"),
                ("דגל", "word"),
                (" ", "whitespace"),
                ("עלה", "word"),
                (" ", "whitespace"),
                ("האדר", "word"),
                ('"', "etc"),
                (" ", "whitespace"),
                ('(', "paren_open"),
                ("או", "word"),
                (" ", "whitespace"),
                ("המייפל", "word"),
                (")", "paren_close"),
                (" ", "whitespace"),
                ("אומץ", "word"),
                (" ", "whitespace"),
                ("בשנת", "word"),
                (" ", "whitespace"),
                ("1965", "number"),
                (" ", "whitespace"),
                ("לאחר", "word"),
                (" ", "whitespace"),
                ("דיון", "word"),
                (" ", "whitespace"),
                ("ציבורי", "word"),
                (" ", "whitespace"),
                ("סביב", "word"),
                (" ", "whitespace"),
                ("שאלת", "word"),
                (" ", "whitespace"),
                ("הסמלים", "word"),
                (" ", "whitespace"),
                ("הלאומיים", "word"),
                (" ", "whitespace"),
                ("שבו", "word"),
                (".", "period")]

    tokens = list(wikitext_split.tokenize(input))

    for token, (s, t) in zip(tokens, expected):
        print(repr(token), (s, t))
        eq_(token, s)
        eq_(token.type, t)


def test_hindi():
    input = 'वसा अर्थात चिकनाई शरीर को क्रियाशील बनाए रखने मे सहयोग करती है।'

    expected = [("वसा", "word"),
                (" ", "whitespace"),
                ("अर्थात", "word"),
                (" ", "whitespace"),
                ("चिकनाई", "word"),
                (" ", "whitespace"),
                ("शरीर", "word"),
                (" ", "whitespace"),
                ("को", "word"),
                (" ", "whitespace"),
                ("क्रियाशील", "word"),
                (" ", "whitespace"),
                ("बनाए", "word"),
                (" ", "whitespace"),
                ("रखने", "word"),
                (" ", "whitespace"),
                ("मे", "word"),
                (" ", "whitespace"),
                ("सहयोग", "word"),
                (" ", "whitespace"),
                ("करती", "word"),
                (" ", "whitespace"),
                ("है", "word"),
                ("।", "danda")]

    tokens = list(wikitext_split.tokenize(input))

    for token, (s, t) in zip(tokens, expected):
        print(repr(token), (s, t))
        eq_(token, s)
        eq_(token.type, t)



__name__ = "deltas"
__version__ = "0.6.0"
__author__ = "Aaron Halfaker"
__author_email__ = "aaron.halfaker@gmail.com"
__description__ = "An experimental diff library for generating " + \
                  "operation deltas that represent the " + \
                  "difference between two sequences of comparable items."
__license__ = "MIT"
__url__ = "https://github.com/halfak/deltas"

all = [__name__, __version__, __author__, __author_email__, __description__,
       __license__, __url__]



"""
The primary use-case of this library is to detect differences between two
sequences of tokens.  So far, two such algorithmic strategies are available:

:class:`~deltas.algorithms.sequence_matcher`
    implementes :func:`~deltas.algorithms.sequence_matcher.diff` that will
    compare two sequences of :class:`~deltas.Token` and return
    a set of operations.
:class:`~deltas.algorithms.segment_matcher`
    implementes :func:`~deltas.algorithms.segment_matcher.diff` that
    uses a :class:`~deltas.Segmenter` to detect block moves

Both of these algorithms are supplimented with a :class:`deltas.DiffEngine`
for efficiently processing several revisions of the same text

Implemented Algorithms
----------------------

Segment Matcher
+++++++++++++++
.. automodule:: deltas.algorithms.segment_matcher

Sequence Matcher
++++++++++++++++
.. automodule:: deltas.algorithms.sequence_matcher

Diff engine
-----------
.. automodule:: deltas.algorithms.diff_engine
"""
from .diff_engine import DiffEngine
from .segment_matcher import SegmentMatcher
from .sequence_matcher import SequenceMatcher


__all__ = [DiffEngine, SegmentMatcher, SequenceMatcher]



from .tokenizer import RegexTokenizer

LEXICON = [
    ('word',          r'[^\W\d]+'),
    ('number',        r'[\d]+'),
    ('period',        r'\.'),
    ('qmark',         r'\?'),
    ('epoint',        r'!'),
    ('comma',         r','),
    ('colon',         r':'),
    ('scolon',        r';'),
    ('break',         r'(\n|\n\r|\r\n)\s*(\n|\n\r|\r\n)+'),
    ('whitespace',    r'[\n\r\s]+'),
    ("etc",           r".")
]

text_split = RegexTokenizer(LEXICON)



import os

from setuptools import find_packages, setup

about_path = os.path.join(os.path.dirname(__file__), "deltas/about.py")
exec(compile(open(about_path).read(), about_path, "exec"))


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


def requirements(fname):
    return [line.strip()
            for line in open(os.path.join(os.path.dirname(__file__), fname))]


setup(
    name=__name__,  # noqa
    version=__version__,  # noqa
    author=__author__,  # noqa
    author_email=__author_email__,  # noqa
    description=__description__,  # noqa
    url=__url__,  # noqa
    license=__license__,  # noqa
    packages=find_packages(),
    long_description=read('README.MD'),
    long_description_content_type="text/markdown",
    install_requires=requirements('requirements.txt'),
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: MIT License",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Topic :: Text Processing :: Linguistic",
        "Topic :: Text Processing :: General",
        "Topic :: Utilities",
        "Topic :: Scientific/Engineering"
    ]
)



from ...tests.diff_and_replay import diff_and_replay
from ...tests.diff_sequence import diff_sequence
from ..sequence_matcher import diff, process


def test_diff_and_replay():
    return diff_and_replay(diff)


def test_engine():
    return diff_sequence(process)



from nose.tools import eq_

from ..wikitext_split_w_cjk import wikitext_split_w_cjk


def test_chinese():
    input = "长岛三百周年半美元是美国铸币局1936年生产的50美分纪念币，" + \
            "旨在纪念欧洲移民定居长岛三百周年。"
    expected = [('长岛', 'cjk_word'),
                ('三百', 'cjk_word'),
                ('周年', 'cjk_word'),
                ('半', 'cjk_word'),
                ('美元', 'cjk_word'),
                ('是', 'cjk_word'),
                ('美国', 'cjk_word'),
                ('铸币', 'cjk_word'),
                ('局', 'cjk_word'),
                ('1936', 'number'),
                ('年', 'cjk_word'),
                ('生产', 'cjk_word'),
                ('的', 'cjk_word'),
                ('50', 'number'),
                ('美分', 'cjk_word'),
                ('纪念币', 'cjk_word'),
                ('，', 'etc'),
                ('旨在', 'cjk_word'),
                ('纪念', 'cjk_word'),
                ('欧洲', 'cjk_word'),
                ('移民', 'cjk_word'),
                ('定居', 'cjk_word'),
                ('长岛', 'cjk_word'),
                ('三百', 'cjk_word'),
                ('周年', 'cjk_word'),
                ('。', 'japan_punct')]

    tokens = list(wikitext_split_w_cjk.tokenize(input))

    for token, (s, t) in zip(tokens, expected):
        print(repr(token), (s, t))
        eq_(token, s)
        eq_(token.type, t)


def test_japanese():
    input = "日本語の俳句（はいく）とは、季語（有季）及び五・七・五（十七音）" + \
            "を主とした定型を基本とする定型詩である。江戸時代には十七文字と呼称され、" + \
            "現代では十七音とも表記される[1]。和歌や連句（俳諧連歌）の発句と同様に、" + \
            "俳句は発生の時点で無季（雑）の作品も存在しており、無季俳句といわれる。有季定型" + \
            "性を捨象する形で派生した自由律俳句もある。また、多くの外国語でも俳句は作られて" + \
            "いるが、外国語では季節感のある言葉でも季語の本意・本情が成立しがたく、しかも五" + \
            "・七・五では切れや季語が活きる短さにならない言語が多いため、日本で言うところの" + \
            "無季自由律俳句が多い。世界最短の定型詩のうちの一つとされる。俳句を詠む（作る）人を俳人と呼ぶ。 "

    expected = [('日本語', 'cjk_word'),
                ('の', 'cjk_word'),
                ('俳句', 'cjk_word'),
                ('（', 'etc'),
                ('は', 'cjk_word'),
                ('いく', 'cjk_word'),
                ('）', 'etc'),
                ('と', 'cjk_word'),
                ('は', 'cjk_word'),
                ('、', 'japan_punct'),
                ('季語', 'cjk_word'),
                ('（', 'etc'),
                ('有季', 'cjk_word'),
                ('）', 'etc'),
                ('及び', 'cjk_word'),
                ('五', 'cjk_word'),
                ('・', 'cjk_word'),
                ('七', 'cjk_word'),
                ('・', 'cjk_word'),
                ('五', 'cjk_word'),
                ('（', 'etc'),
                ('十七', 'cjk_word'),
                ('音', 'cjk_word'),
                ('）', 'etc'),
                ('を', 'cjk_word'),
                ('主', 'cjk_word'),
                ('と', 'cjk_word'),
                ('し', 'cjk_word'),
                ('た', 'cjk_word'),
                ('定型', 'cjk_word'),
                ('を', 'cjk_word'),
                ('基本', 'cjk_word'),
                ('と', 'cjk_word'),
                ('する', 'cjk_word'),
                ('定型詩', 'cjk_word'),
                ('で', 'cjk_word'),
                ('ある', 'cjk_word'),
                ('。', 'japan_punct'),
                ('江戸', 'cjk_word'),
                ('時代', 'cjk_word'),
                ('に', 'cjk_word'),
                ('は', 'cjk_word'),
                ('十七', 'cjk_word'),
                ('文字', 'cjk_word'),
                ('と', 'cjk_word'),
                ('呼称', 'cjk_word'),
                ('さ', 'cjk_word'),
                ('れ', 'cjk_word'),
                ('、', 'japan_punct'),
                ('現代', 'cjk_word'),
                ('で', 'cjk_word'),
                ('は', 'cjk_word'),
                ('十七', 'cjk_word'),
                ('音', 'cjk_word'),
                ('とも', 'cjk_word'),
                ('表記', 'cjk_word'),
                ('さ', 'cjk_word'),
                ('れる', 'cjk_word'),
                ('[', 'brack_open'),
                ('1', 'number'),
                (']', 'brack_close'),
                ('。', 'japan_punct'),
                ('和歌', 'cjk_word'),
                ('や', 'cjk_word'),
                ('連句', 'cjk_word'),
                ('（', 'etc'),
                ('俳諧', 'cjk_word'),
                ('連歌', 'cjk_word'),
                ('）', 'etc'),
                ('の', 'cjk_word'),
                ('発句', 'cjk_word'),
                ('と', 'cjk_word'),
                ('同様', 'cjk_word'),
                ('に', 'cjk_word'),
                ('、', 'japan_punct'),
                ('俳句', 'cjk_word'),
                ('は', 'cjk_word'),
                ('発生', 'cjk_word'),
                ('の', 'cjk_word'),
                ('時点', 'cjk_word'),
                ('で', 'cjk_word'),
                ('無季', 'cjk_word'),
                ('（', 'etc'),
                ('雑', 'cjk_word'),
                ('）', 'etc'),
                ('の', 'cjk_word'),
                ('作品', 'cjk_word'),
                ('も', 'cjk_word'),
                ('存在', 'cjk_word'),
                ('し', 'cjk_word'),
                ('て', 'cjk_word'),
                ('おり', 'cjk_word'),
                ('、', 'japan_punct'),
                ('無季', 'cjk_word'),
                ('俳句', 'cjk_word'),
                ('と', 'cjk_word'),
                ('いわ', 'cjk_word'),
                ('れる', 'cjk_word'),
                ('。', 'japan_punct'),
                ('有季', 'cjk_word'),
                ('定型', 'cjk_word'),
                ('性', 'cjk_word'),
                ('を', 'cjk_word'),
                ('捨象', 'cjk_word'),
                ('する', 'cjk_word'),
                ('形', 'cjk_word'),
                ('で', 'cjk_word'),
                ('派生', 'cjk_word'),
                ('し', 'cjk_word'),
                ('た', 'cjk_word'),
                ('自由', 'cjk_word'),
                ('律', 'cjk_word'),
                ('俳句', 'cjk_word'),
                ('も', 'cjk_word'),
                ('ある', 'cjk_word'),
                ('。', 'japan_punct'),
                ('また', 'cjk_word'),
                ('、', 'japan_punct'),
                ('多く', 'cjk_word'),
                ('の', 'cjk_word'),
                ('外国語', 'cjk_word'),
                ('で', 'cjk_word'),
                ('も', 'cjk_word'),
                ('俳句', 'cjk_word'),
                ('は', 'cjk_word'),
                ('作ら', 'cjk_word'),
                ('れ', 'cjk_word'),
                ('て', 'cjk_word'),
                ('いる', 'cjk_word'),
                ('が', 'cjk_word'),
                ('、', 'japan_punct'),
                ('外国語', 'cjk_word'),
                ('で', 'cjk_word'),
                ('は', 'cjk_word'),
                ('季節感', 'cjk_word'),
                ('の', 'cjk_word'),
                ('ある', 'cjk_word'),
                ('言葉', 'cjk_word'),
                ('で', 'cjk_word'),
                ('も', 'cjk_word'),
                ('季語', 'cjk_word'),
                ('の', 'cjk_word'),
                ('本意', 'cjk_word'),
                ('・', 'cjk_word'),
                ('本', 'cjk_word'),
                ('情', 'cjk_word'),
                ('が', 'cjk_word'),
                ('成立', 'cjk_word'),
                ('し', 'cjk_word'),
                ('がたく', 'cjk_word'),
                ('、', 'japan_punct'),
                ('しかも', 'cjk_word'),
                ('五', 'cjk_word'),
                ('・', 'cjk_word'),
                ('七', 'cjk_word'),
                ('・', 'cjk_word'),
                ('五', 'cjk_word'),
                ('で', 'cjk_word'),
                ('は', 'cjk_word'),
                ('切れ', 'cjk_word'),
                ('や', 'cjk_word'),
                ('季語', 'cjk_word'),
                ('が', 'cjk_word'),
                ('活きる', 'cjk_word'),
                ('短', 'cjk_word'),
                ('さ', 'cjk_word'),
                ('に', 'cjk_word'),
                ('なら', 'cjk_word'),
                ('ない', 'cjk_word'),
                ('言語', 'cjk_word'),
                ('が', 'cjk_word'),
                ('多い', 'cjk_word'),
                ('ため', 'cjk_word'),
                ('、', 'japan_punct'),
                ('日本', 'cjk_word'),
                ('で', 'cjk_word'),
                ('言う', 'cjk_word'),
                ('ところ', 'cjk_word'),
                ('の', 'cjk_word'),
                ('無季', 'cjk_word'),
                ('自由', 'cjk_word'),
                ('律', 'cjk_word'),
                ('俳句', 'cjk_word'),
                ('が', 'cjk_word'),
                ('多い', 'cjk_word'),
                ('。', 'japan_punct'),
                ('世界', 'cjk_word'),
                ('最短', 'cjk_word'),
                ('の', 'cjk_word'),
                ('定型詩', 'cjk_word'),
                ('の', 'cjk_word'),
                ('うち', 'cjk_word'),
                ('の', 'cjk_word'),
                ('一', 'cjk_word'),
                ('つ', 'cjk_word'),
                ('と', 'cjk_word'),
                ('さ', 'cjk_word'),
                ('れる', 'cjk_word'),
                ('。', 'japan_punct'),
                ('俳句', 'cjk_word'),
                ('を', 'cjk_word'),
                ('詠む', 'cjk_word'),
                ('（', 'etc'),
                ('作る', 'cjk_word'),
                ('）', 'etc'),
                ('人', 'cjk_word'),
                ('を', 'cjk_word'),
                ('俳人', 'cjk_word'),
                ('と', 'cjk_word'),
                ('呼ぶ', 'cjk_word'),
                ('。', 'japan_punct'),
                (' ', 'whitespace')]

    tokens = list(wikitext_split_w_cjk.tokenize(input))

    for token, (s, t) in zip(tokens, expected):
        print(repr(token), (s, t))
        eq_(token, s)
        eq_(token.type, t)


def test_korean():
    input = "국어사 자료에서 ‘김치’가 소급하는 최초의 형태는 16세기의 " + \
            "‘딤치’이다. 이 단어는 한자어 ‘침채(沈菜)’에서 온 말인데, ‘딤치’는 " + \
            "16세기보는 훨씬 이전의 한자음을 반영한 것이다. 17세기에 나타나는 " + \
            "‘짐치’는 이례적인데, 아마 경상 방언형이 아닌가 한다. 왜냐하면 서울 말에서 " + \
            "‘디’가 ‘지’로 바뀐 구개음화는 17세기 말에서 18세기초에 일반적으로 일어났기 때문이다. "

    expected = [('국어', 'cjk_word'),
                ('사', 'cjk_word'),
                (' ', 'whitespace'),
                ('자료', 'cjk_word'),
                (' ', 'whitespace'),
                ('‘', 'etc'),
                ('김치', 'cjk_word'),
                (' ', 'whitespace'),
                ('소급', 'cjk_word'),
                (' ', 'whitespace'),
                ('최초', 'cjk_word'),
                (' ', 'whitespace'),
                ('형태', 'cjk_word'),
                (' ', 'whitespace'),
                ('16', 'number'),
                ('세기', 'cjk_word'),
                (' ', 'whitespace'),
                ('‘', 'etc'),
                ('딤치', 'cjk_word'),
                ('.', 'period'),
                (' ', 'whitespace'),
                ('이', 'cjk_word'),
                (' ', 'whitespace'),
                ('단어', 'cjk_word'),
                (' ', 'whitespace'),
                ('한자어', 'cjk_word'),
                (' ', 'whitespace'),
                ('‘', 'etc'),
                ('침채', 'cjk_word'),
                ('(', 'paren_open'),
                ('沈菜', 'cjk_word'),
                (')', 'paren_close'),
                ('’', 'etc'),
                ('에서', 'cjk_word'),
                (' ', 'whitespace'),
                ('온', 'cjk_word'),
                (' ', 'whitespace'),
                ('말', 'cjk_word'),
                (',', 'comma'),
                (' ', 'whitespace'),
                ('‘', 'etc'),
                ('딤치', 'cjk_word'),
                (' ', 'whitespace'),
                ('16', 'number'),
                ('기보', 'cjk_word'),
                (' ', 'whitespace'),
                ('훨씬', 'cjk_word'),
                (' ', 'whitespace'),
                ('이전', 'cjk_word'),
                (' ', 'whitespace'),
                ('자음', 'cjk_word'),
                (' ', 'whitespace'),
                ('반영', 'cjk_word'),
                (' ', 'whitespace'),
                ('것', 'cjk_word'),
                ('.', 'period'),
                (' ', 'whitespace'),
                ('17', 'number'),
                ('세기', 'cjk_word'),
                (' ', 'whitespace'),
                ('나타나는', 'cjk_word'),
                (' ', 'whitespace'),
                ('‘', 'etc'),
                ('짐치’는', 'cjk_word'),
                (' ', 'whitespace'),
                ('이례', 'cjk_word'),
                (',', 'comma'),
                (' ', 'whitespace'),
                ('아마', 'cjk_word'),
                (' ', 'whitespace'),
                ('경상', 'cjk_word'),
                (' ', 'whitespace'),
                ('방언', 'cjk_word'),
                (' ', 'whitespace'),
                ('아닌가', 'cjk_word'),
                (' ', 'whitespace'),
                ('한다', 'cjk_word'),
                ('.', 'period'),
                (' ', 'whitespace'),
                ('왜냐하면', 'cjk_word'),
                (' ', 'whitespace'),
                ('서울', 'cjk_word'),
                (' ', 'whitespace'),
                ('말', 'cjk_word'),
                (' ', 'whitespace'),
                ('‘', 'etc'),
                ('디', 'cjk_word'),
                (' ', 'whitespace'),
                ('‘', 'etc'),
                ('로', 'cjk_word'),
                (' ', 'whitespace'),
                ('바뀐', 'cjk_word'),
                (' ', 'whitespace'),
                ('구개음화', 'cjk_word'),
                (' ', 'whitespace'),
                ('17', 'number'),
                ('세기', 'cjk_word'),
                (' ', 'whitespace'),
                ('말', 'cjk_word'),
                (' ', 'whitespace'),
                ('18', 'number'),
                ('기초', 'cjk_word'),
                (' ', 'whitespace'),
                ('일반', 'cjk_word'),
                (' ', 'whitespace'),
                ('일어났기', 'cjk_word'),
                (' ', 'whitespace'),
                ('때문', 'cjk_word'),
                ('.', 'period'),
                (' ', 'whitespace')]

    tokens = list(wikitext_split_w_cjk.tokenize(input))

    for token, (s, t) in zip(tokens, expected):
        print(repr(token), (s, t))
        eq_(token, s)
        eq_(token.type, t)
